
from OpenSSL import crypto
from anynet import util
import contextlib
import tempfile
import anyio
import ssl
import os

import logging
logger = logging.getLogger(__name__)


TYPE_DER = 0
TYPE_PEM = 1

VERSION_TLS = 0
VERSION_TLS11 = 1
VERSION_TLS12 = 2

TypeMap = {
	TYPE_DER: crypto.FILETYPE_ASN1,
	TYPE_PEM: crypto.FILETYPE_PEM
}

VersionMap = {
	VERSION_TLS: ssl.PROTOCOL_TLS,
	VERSION_TLS11: ssl.PROTOCOL_TLSv1_1,
	VERSION_TLS12: ssl.PROTOCOL_TLSv1_2
}


def set_certificate_chain(context, certs, key):
	certfile = tempfile.NamedTemporaryFile(delete=False)
	keyfile = tempfile.NamedTemporaryFile(delete=False)
	
	for cert in certs:
		certfile.write(cert.encode(TYPE_PEM))
	keyfile.write(key.encode(TYPE_PEM))
	
	certfile.close()
	keyfile.close()
	
	context.load_cert_chain(certfile.name, keyfile.name)
	
	os.remove(certfile.name)
	os.remove(keyfile.name)

def load_certificate_chain(filename):
	with open(filename, "rb") as f:
		data = f.read()
	
	header = b"-----BEGIN CERTIFICATE-----"
	
	certs = []
	for part in data.split(header)[1:]:
		certs.append(TLSCertificate.parse(header + part, TYPE_PEM))
	return certs


class X509Name:
	ITEMS = {
		"C": "country_name",
		"ST": "state_or_province_name",
		"L": "locality_name",
		"O": "organization_name",
		"OU": "organizational_unit_name",
		"CN": "common_name",
		"E": "email_address"
	}
	
	ATTRS = {
		"country_name": "countryName",
		"state_or_province_name": "stateOrProvinceName",
		"locality_name": "localityName",
		"organization_name": "organizationName",
		"organizational_unit_name": "organizationalUnitName",
		"common_name": "commonName",
		"email_address": "emailAddress"
	}
	
	def __init__(self):
		self.obj = crypto.X509().get_subject()
	
	def __getitem__(self, key):
		return getattr(self, self.ITEMS[key])
	def __setitem__(self, key, value):
		setattr(self, self.ITEMS[key], value)
		
	def __getattr__(self, name):
		return getattr(self.__dict__["obj"], X509Name.ATTRS[name])
	def __setattr__(self, name, value):
		if name == "obj":
			self.__dict__["obj"] = value
		else:
			setattr(self.obj, X509Name.ATTRS[name], value)


class TLSCertificate:
	def __init__(self, obj):
		self.obj = obj
		self.subject = X509Name()
		self.subject.obj = obj.get_subject()
		self.issuer = X509Name()
		self.issuer.obj = obj.get_issuer()
		
	def public_key(self):
		pkey = self.obj.get_pubkey()
		rsakey = pkey.to_cryptography_key()
		return rsakey.public_numbers()
		
	def encode(self, format):
		return crypto.dump_certificate(TypeMap[format], self.obj)
		
	def save(self, filename, format):
		with open(filename, "wb") as f:
			f.write(self.encode(format))
		
	def sign(self, key, alg="sha256"):
		self.obj.sign(key.obj, alg)
	
	@classmethod
	def load(cls, filename, format):
		with open(filename, "rb") as f:
			data = f.read()
		return cls.parse(data, format)
		
	@classmethod
	def parse(cls, data, format):
		cert = crypto.load_certificate(TypeMap[format], data)
		return cls(cert)
		
	@classmethod
	def generate(cls, key):
		cert = crypto.X509()
		cert.set_pubkey(key.obj)
		
		cert.set_notBefore(b"20000101000000Z")
		cert.set_notAfter(b"29990101000000Z")
		
		return cls(cert)

	
class TLSPrivateKey:
	def __init__(self, obj):
		self.obj = obj
		
	def encode(self, format):
		return crypto.dump_privatekey(TypeMap[format], self.obj)
	
	def save(self, filename, format):
		with open(filename, "wb") as f:
			f.write(self.encode(format))
	
	@classmethod
	def load(cls, filename, format):
		with open(filename, "rb") as f:
			data = f.read()
		return cls.parse(data, format)
		
	@classmethod
	def parse(cls, data, format):
		pkey = crypto.load_privatekey(TypeMap[format], data)
		return cls(pkey)
		
	@classmethod
	def generate(cls, size=2048):
		pkey = crypto.PKey()
		pkey.generate_key(crypto.TYPE_RSA, size)
		return cls(pkey)


class TLSContext:
	def __init__(self, version=VERSION_TLS):
		self.version = version
		self.authority = None
		self.certs = None
		self.key = None
	
	def set_certificate(self, cert, key):
		self.certs = [cert]
		self.key = key
	
	def set_certificate_chain(self, certs, key):
		self.certs = certs
		self.key = key
	
	def set_authority(self, authority):
		self.authority = authority
	
	def get(self, server):
		if server:
			return self.make_server_context()
		return self.make_client_context()
	
	def make_server_context(self):
		context = ssl.SSLContext(VersionMap[self.version])
		if self.certs and self.key:
			set_certificate_chain(context, self.certs, self.key)
		else:
			raise ValueError("Please provide a server certificate")
		
		context.verify_mode = ssl.CERT_NONE
		if self.authority:
			data = self.authority.encode(TYPE_DER)
			context.load_verify_locations(cadata=data)
			context.verify_mode = ssl.CERT_REQUIRED

		context.check_hostname = False
		return context
	
	def make_client_context(self):
		context = ssl.SSLContext(VersionMap[self.version])
		if self.certs and self.key:
			set_certificate_chain(context, self.certs, self.key)
		
		context.verify_mode = ssl.CERT_REQUIRED
		if self.authority:
			data = self.authority.encode(TYPE_DER)
			context.load_verify_locations(cadata=data)
		else:
			context.load_default_certs()
		
		context.check_hostname = True
		return context


class TLSClient:
	def __init__(self, stream):
		self.stream = stream
		self.lock = anyio.Lock()
		
	async def send(self, data):
		async with self.lock:
			await self.stream.send(data)
	async def recv(self, num=65536):
		return await self.stream.receive(num)
	
	def local_address(self):
		return self.stream.extra(anyio.abc.SocketAttribute.local_address)
	def remote_address(self):
		return self.stream.extra(anyio.abc.SocketAttribute.remote_address)
	
	def remote_certificate(self):
		cert = self.stream.extra(anyio.streams.tls.TLSAttribute.peer_certificate_binary, None)
		if cert:
			return TLSCertificate.parse(cert, TYPE_DER)


@contextlib.asynccontextmanager
async def connect(host, port, context=None):
	logger.debug("Connecting TLS client to %s:%s", host, port)
	if context:
		context = context.get(False)
	async with await anyio.connect_tcp(host, port, ssl_context=context, tls_standard_compatible=False) as stream:
		yield TLSClient(stream)

@contextlib.asynccontextmanager
async def create_listener(host, port, context):
	listener = await anyio.create_tcp_listener(local_host=host, local_port=port)
	async with listener:
		if context:
			listener = anyio.streams.tls.TLSListener(listener, context.get(True), False)
			async with listener:
				yield listener
		else:
			yield listener

@contextlib.asynccontextmanager
async def serve(handler, host="", port=0, context=None):
	async def handle(stream):
		with util.catch(Exception):
			if context:
				try:
					stream = await anyio.streams.tls.TLSStream.wrap(
						stream, ssl_context=context.get(True), standard_compatible=False
					)
				except util.StreamError:
					logger.warning("Failed to accept TLS handshake")
					return
			
			async with stream:
				await handler(TLSClient(stream))
				logger.debug("Closing TLS connection")
	
	if not host:
		host = util.local_address()
	
	logger.info("Starting TLS server at %s:%i", host, port)
	async with await anyio.create_tcp_listener(local_host=host, local_port=port) as listener:
		async with util.create_task_group() as group:
			group.start_soon(listener.serve, handle)
			yield




import math
import anyio


class Queue:
	def __init__(self, stream):
		self.stream = stream
	
	async def __aenter__(self): return self
	async def __aexit__(self, typ, val, tb):
		await self.stream.aclose()
	
	async def put(self, value):
		await self.stream.send(value)
	
	async def get(self):
		return await self.stream.receive()
	
	async def close(self):
		await self.stream.aclose()


def create(size=math.inf):
	send, recv = anyio.create_memory_object_stream(size)
	stream = anyio.streams.stapled.StapledObjectStream(send, recv)
	return Queue(stream)




class RC4:
	def __init__(self, key, reset=False):
		self.reinit = reset
		self.set_key(key)
	
	def set_key(self, key):
		self.key = key
		self.reset()
		
	def reset(self):
		self.state = list(range(256))
		self.x = self.y = 0
		
		p = 0
		for i in range(256):
			p = (p + self.state[i] + self.key[i % len(self.key)]) % 256
			self.state[i], self.state[p] = self.state[p], self.state[i]
		
	def crypt(self, data):
		out = []
		for char in data:
			self.x = (self.x + 1) % 256
			self.y = (self.y + self.state[self.x]) % 256
			p, q = self.state[self.y], self.state[self.x]
			self.state[self.x], self.state[self.y] = p, q
			
			mask = self.state[(p + q) % 256]
			out.append(char ^ mask)
			
		if self.reinit:
			self.reset()

		return bytes(out)




from anynet import util
import contextlib
import socket
import anyio

import logging
logger = logging.getLogger(__name__)


class TCPClient:
	def __init__(self, stream):
		self.stream = stream
		self.lock = anyio.Lock()
	
	async def send(self, data):
		async with self.lock:
			await self.stream.send(data)
	async def recv(self, num=65536):
		return await self.stream.receive(num)
	
	def local_address(self):
		return self.stream.extra(anyio.abc.SocketAttribute.local_address)
	def remote_address(self):
		return self.stream.extra(anyio.abc.SocketAttribute.remote_address)


@contextlib.asynccontextmanager
async def connect(host, port):
	logger.debug("Connecting TCP client to %s:%i", host, port)
	async with await anyio.connect_tcp(host, port) as stream:
		yield TCPClient(stream)

@contextlib.asynccontextmanager
async def serve(handler, host="", port=0):
	async def handle(stream):
		async with stream:
			with util.catch(Exception):
				await handler(TCPClient(stream))
		
	if not host:
		host = util.local_address()
	
	logger.info("Starting TCP server at %s:%i", host, port)
	
	listener = await anyio.create_tcp_listener(local_host=host, local_port=port)
	async with listener:
		async with util.create_task_group() as group:
			group.start_soon(listener.serve, handle)
			yield




import string


NAME_CHARS = string.ascii_letters + string.digits + ":-_"


def decode_entities(s):
	s = s.replace("&quot;", '"')
	s = s.replace("&apos;", "'")
	s = s.replace("&lt;", "<")
	s = s.replace("&gt;", ">")
	s = s.replace("&amp;", "&")
	return s

def encode_entities(s):
	s = s.replace("&", "&amp;")
	s = s.replace("'", "&quot;")
	s = s.replace('"', "&apos;")
	s = s.replace("<", "&lt;")
	s = s.replace(">", "&gt;")
	return s


class TextStream:
	def __init__(self, text):
		self.text = text
		self.pos = 0
		
	def peek(self, size=1):
		if self.available() < size:
			raise OverflowError("Buffer overflow in text stream")
		return self.text[self.pos : self.pos + size]
		
	def read(self, size=1):
		if self.available() < size:
			raise OverflowError("Buffer overflow in text stream")
		text = self.text[self.pos : self.pos + size]
		self.pos += size
		return text
		
	def skip(self, size=1):
		self.pos += size
		
	def available(self): return len(self.text) - self.pos
		
	def eof(self):
		return self.pos == len(self.text)
		
	def skip_whitespace(self):
		while not self.eof():
			char = self.peek()
			if char not in string.whitespace:
				return
			self.skip(1)


class XMLTree:
	def __init__(self, name):
		self.children = []
		self.attrs = {}
		
		self.text = None
		self.name = name
	
	def __str__(self):
		return self.encode()
	
	def __contains__(self, name):
		for node in self.children:
			if node.name == name:
				return True
		return False
	
	def __getitem__(self, name):
		for node in self.children:
			if node.name == name:
				return node
		raise KeyError(name)
		
	def __iter__(self):
		return iter(self.children)
	
	def __len__(self):
		return len(self.children)
		
	def find(self, name):
		nodes = []
		for node in self.children:
			if node.name == name:
				nodes.append(node)
		return nodes
	
	def add(self, name, text=None, attrs={}):
		node = XMLTree(name)
		node.text = text
		node.attrs = dict(attrs)
		self.children.append(node)
		return node
	
	def encode(self):
		data = "<%s" %self.name
		for name, value in self.attrs.items():
			data += ' %s="%s"' %(name, encode_entities(value))
		data += ">"
		
		for child in self.children:
			data += child.encode()

		if self.text is not None:
			data += encode_entities(str(self.text))
		
		data += "</%s>" %self.name
		return data
		
		
class XMLParser:
	def parse(self, text):
		stream = TextStream(text)
		self.parse_declaration(stream)
		
		stream.skip_whitespace()
		
		tree = self.parse_tree(stream)
		
		stream.skip_whitespace()
		if not stream.eof():
			raise ValueError("XML document has data behind root tag")
		return tree
		
	def parse_declaration(self, stream):
		if stream.peek(6) == "<?xml ":
			stream.read(6)
			
			self.parse_declaration_attribs(stream)
			
			if stream.read() != "?":
				raise ValueError("XML declaration is invalid")
			stream.skip_whitespace()
			
			if stream.read() != ">":
				raise ValueError("XML declaration is invalid")
			
	def parse_declaration_attribs(self, stream):
		version = self.parse_fixed_attribute(stream, "version")
		if version != "1.0":
			raise ValueError("XML version must be 1.0")
		if stream.peek() == "?": return
		
		encoding = self.parse_fixed_attribute(stream, "encoding")
		if stream.peek() == "?": return
		
		standalone = self.parse_fixed_attribute(stream, "standalone")
		if standalone not in ["yes", "no"]:
			raise ValueError("standalone must be either yes of no")
			
	def parse_tree(self, stream):
		if stream.read() != "<":
			raise ValueError("Unexpected character in XML document")
			
		stream.skip_whitespace()
		
		name = self.parse_name(stream)
		tree = XMLTree(name)
		
		stream.skip_whitespace()
		
		char = stream.peek()
		while char not in "/>":
			name, value = self.parse_attribute(stream)
			if name in tree.attrs:
				raise ValueError("Duplicate attributein XML document")
			tree.attrs[name] = value
			char = stream.peek()
		
		char = stream.read()
		if char == "/":
			if stream.read() != ">":
				raise ValueError("Unexpected character in XML document")
			return tree
			
		tree.text = ""
		
		chars = stream.peek(2)
		while chars != "</":
			if chars[0] == "<":
				tree.children.append(self.parse_tree(stream))
			else:
				tree.text += chars[0]
				stream.skip()
			chars = stream.peek(2)
			
		tree.text = decode_entities(tree.text)
			
		stream.skip(2)
		stream.skip_whitespace()
		
		name = self.parse_name(stream)
		if name != tree.name:
			raise ValueError(
				"Closing tag has unexpected name: '%s' (expected '%s')" %(name, tree.name)
			)
			
		stream.skip_whitespace()
		if stream.read() != ">":
			raise ValueError("Unexpected character in XML document")
		
		return tree	
			
	def parse_fixed_attribute(self, stream, attr):
		name, value = self.parse_attribute(stream)
		if name != attr:
			raise ValueError("Expected '%s' attribute, not '%s'" %(attr, name))
		return value
		
	def parse_attribute(self, stream):
		stream.skip_whitespace()
		key = self.parse_name(stream)
		stream.skip_whitespace()
		if stream.read() != "=":
			raise ValueError("Expected '=' after attribute name")
		stream.skip_whitespace()
		value = self.parse_string(stream)
		stream.skip_whitespace()
		return key, value
		
	def parse_string(self, stream):
		strchar = stream.read()
		if strchar not in ["'", '"']:
			raise ValueError("Expected string attribute")
		
		string = ""
		char = stream.read()
		while char != strchar:
			string += char
			char = stream.read()
		
		string = " ".join(string.split())
		string = decode_entities(string)
		return string
		
	def parse_name(self, stream):
		name = ""
		char = stream.peek()
		while char in NAME_CHARS:
			name += char
			stream.skip()
			char = stream.peek()
		return name
		
		
def parse(text):
	parser = XMLParser()
	try:
		return parser.parse(text)
	except OverflowError:
		raise ValueError("XML document is incomplete")




from anynet import tls, util, xml, scheduler
from multidict import MultiDict, CIMultiDict
import urllib.parse
import contextlib
import datetime
import anyio
import json

import logging
logger = logging.getLogger(__name__)


STATUS_NAMES = {
	100: "Continue",
	101: "Switching Protocols",
	102: "Processing",
	103: "Early Hints",
	200: "OK",
	201: "Created",
	202: "Accepted",
	203: "Non-Authoritative Information",
	204: "No Content",
	205: "Reset Content",
	206: "Partial Content",
	207: "Multi-Status",
	208: "Already Reported",
	226: "IM Used",
	300: "Multiple Choices",
	301: "Moved Permanently",
	302: "Found",
	303: "See Other",
	304: "Not Modified",
	305: "Use Proxy",
	307: "Temporary Redirect",
	308: "Permanent Redirect",
	400: "Bad Request",
	401: "Unauthorized",
	402: "Payment Required",
	403: "Forbidden",
	404: "Not Found",
	405: "Method Not Allowed",
	406: "Not Acceptable",
	407: "Proxy Authentication Required",
	408: "Request Timeout",
	409: "Conflict",
	410: "Gone",
	411: "Length Required",
	412: "Precondition Failed",
	413: "Payload Too Large",
	414: "URI Too Long",
	415: "Unsupported Media Type",
	416: "Range Not Satisfiable",
	417: "Expectation Failed",
	421: "Misdirected Request",
	422: "Unprocessable Content",
	423: "Locked",
	424: "Failed Dependency",
	425: "Too Early",
	426: "Upgrade Required",
	428: "Precondition Required",
	429: "Too Many Requests",
	431: "Request Header Fields Too Large",
	451: "Unavailable For Legal Reasons",
	500: "Internal Server Error",
	501: "Not Implemented",
	502: "Bad Gateway",
	503: "Service Unavailable",
	504: "Gateway Timeout",
	505: "HTTP Version Not Supported",
	506: "Variant Also Negotiates",
	507: "Insufficient Storage",
	508: "Loop Detected",
	510: "Not Extended",
	511: "Network Authentication Required"
}


XML_TYPES = [
	"application/xml",
	"text/xml"
]

TEXT_TYPES = [
	"application/json",
	"application/x-www-form-urlencoded",
	"application/xml"
]


class HTTPError(Exception): pass

class HTTPResponseError(HTTPError):
	def __init__(self, response):
		self.response = response
	
	def __str__(self):
		return "HTTP request failed: %i (%s)" %(self.response.status_code, self.response.status_name)


def parse_date(text):
	dt = datetime.datetime.strptime(text, "%a, %d %b %Y %H:%M:%S GMT")
	return dt.replace(tzinfo=datetime.timezone.utc)

def format_date(date):
	date = date.astimezone(datetime.timezone.utc)
	return date.strftime("%a, %d %b %Y %H:%M:%S GMT")

def current_date():
	return format_date(datetime.datetime.now())


def urlencode(data):
	return urllib.parse.quote(data)
def urldecode(data):
	return urllib.parse.unquote(data)

def formencode(data, url=True):
	fields = []
	for key, value in data.items():
		if value is None:
			if url:
				key = urlencode(str(key))
			fields.append(key)
		else:
			if url:
				key = urlencode(str(key))
				value = urlencode(str(value))
			fields.append("%s=%s" %(key, value))
	return "&".join(fields)

def formdecode(data, url=True):
	if not data: return MultiDict()
	
	fields = MultiDict()
	for field in data.split("&"):
		if "=" in field:
			key, value = field.split("=", 1)
			if url:
				key = urldecode(key)
				value = urldecode(value)
			fields[key] = value
		else:
			key = field
			if url:
				key = urldecode(key)
			fields[key] = None
	return fields


def parseheader(header):
	fields = header.split(";")
	type = fields[0].strip()
	
	param = {}
	for field in fields[1:]:
		field = field.strip()
		if not "=" in field:
			raise HTTPError("Malformed parameter in header")
		
		key, value = field.split("=", 1)
		if value.startswith('"'):
			if not value.endswith('"'):
				raise HTTPError("Malformed parameter in header")
			value = value[1:-1]
		
		param[key] = value
	return type, param


class HTTPMessage:
	def __init__(self):
		self.version = "HTTP/1.1"
	
		self.headers = CIMultiDict()
		self.body = b""
		
		self.rawform = None
		self.form = None
		self.json = None
		self.xml = None
		self.files = None
		self.text = None
		
		self.boundary = "--------BOUNDARY--------"
		
	def check_version(self):
		if not self.version.startswith("HTTP/"):
			raise HTTPError("HTTP version must start with HTTP/")
		if self.version not in ["HTTP/1.0", "HTTP/1.1"]:
			raise HTTPError("HTTP version not supported")
	
	def transfer_encodings(self):
		encoding = self.headers.get("Transfer-Encoding", "identity")
		return [enc.strip() for enc in encoding.split(",")]
	
	def is_chunked(self):
		return "chunked" in self.transfer_encodings()
		
	def parse_body(self):
		type, param = parseheader(self.headers.get("Content-Type", ""))
		
		is_json = type == "application/json" or type.endswith("+json")
		is_xml = type in XML_TYPES or type.endswith("+xml")
		is_text = type in TEXT_TYPES or type.startswith("text/") or is_json or is_xml
		
		if is_text:
			try:
				self.text = self.body.decode(param.get("charset", "UTF-8"))
			except UnicodeDecodeError:
				raise HTTPError("Failed to decode HTTP body")
		
		if type == "application/x-www-form-urlencoded":
			self.form = formdecode(self.text)
			self.rawform = formdecode(self.text, False)
		
		if is_json:
			try:
				self.json = json.loads(self.text)
			except json.JSONDecodeError:
				raise HTTPError("Failed to decode JSON body")
		
		if is_xml:
			try:
				self.xml = xml.parse(self.text)
			except ValueError as e:
				raise HTTPError("Failed to decode XML body: %s" %e)
		
		if type.startswith("multipart/form-data"):
			if "boundary" not in param:
				raise HTTPError("multipart/form-data required boundary parameter")
			self.boundary = param["boundary"]
			self.files = self.parse_files(self.body)
	
	def parse_files(self, data):
		split = b"--%s" %self.boundary.encode()
		parts = data.split(split)
		
		if parts[-1] != b"--\r\n" or parts[0] != b"":
			raise HTTPError("Failed to decode multipart body")
		
		files = MultiDict()
		for part in parts[1:-1]:
			if part[:2] != b"\r\n" or part[-2:] != b"\r\n":
				raise HTTPError("Failed to decode multipart body")
			part = part[2:-2]
			
			if not b"\r\n\r\n" in part:
				raise HTTPError("Failed to decode multipart body")
			
			head, body = part.split(b"\r\n\r\n", 1)
			try:
				lines = head.decode().split("\r\n")
			except UnicodeDecodeError:
				raise HTTPError("Failed to decode multipart body")
			
			headers = {}
			for header in lines:
				if not ": " in header:
					raise HTTPError("Invalid line in multipart headers")
				key, value = header.split(": ", 1)
				headers[key] = value
			
			if "Content-Disposition" not in headers:
				raise HTTPError("Expected Content-Disposition header in multipart data")
				
			type, param = parseheader(headers["Content-Disposition"])
			if type != "form-data":
				raise HTTPError("Expected form-data header in multipart data")
			
			if "name" not in param:
				raise HTTPError("Expected name parameter in Content-Disposition header")
			files[param["name"]] = body
				
		return files
		
	def encode_body(self):
		text = self.text
		body = self.body
		
		if self.rawform is not None:
			if "Content-Type" not in self.headers:
				self.headers["Content-Type"] = "application/x-www-form-urlencoded"
			text = formencode(self.rawform, False)
		
		elif self.form is not None:
			if "Content-Type" not in self.headers:
				self.headers["Content-Type"] = "application/x-www-form-urlencoded"
			text = formencode(self.form)
		
		elif self.json is not None:
			if "Content-Type" not in self.headers:
				self.headers["Content-Type"] = "application/json"
			text = json.dumps(self.json)
			
		elif self.xml is not None:
			if "Content-Type" not in self.headers:
				self.headers["Content-Type"] = "application/xml"
			text = self.xml.encode()
			
		elif self.files is not None:
			if "Content-Type" not in self.headers:
				self.headers["Content-Type"] = "multipart/form-data"
			self.headers["Content-Type"] += "; boundary=%s" %self.boundary
			
			text = None
			body = b""
			for name, data in self.files.items():
				name = name.replace('"', '\\"')
				body += b"--%s\r\n" %self.boundary.encode()
				body += b"Content-Disposition: form-data; name=\"%s\"\r\n\r\n" %name.encode()
				body += data + b"\r\n"
			body += b"--%s--\r\n" %self.boundary.encode()
		
		if text is not None:
			if "Content-Type" not in self.headers:
				self.headers["Content-Type"] = "text/plain"
			body = text.encode()
		
		if body and "Content-Type" not in self.headers:
			self.headers["Content-Type"] = "application/octet-stream"
		
		if self.is_chunked():
			if not body:
				return b"0\r\n\r\n"
			return b"%x\r\n" %len(body) + body + b"\r\n0\r\n\r\n"
		else:
			if body:
				self.headers["Content-Length"] = len(body)
			return body
	
	def encode_start_line(self):
		raise NotImplementedError("%s.encode_start_line" %self.__class__.__name__)
	
	def encode_headers(self):
		self.encode_body()
		
		lines = [self.encode_start_line()]
		for key, value in self.headers.items():
			lines.append("%s: %s" %(key, value))
		
		text = "\r\n".join(lines) + "\r\n\r\n"
		return text.encode()
	
	def encode(self):
		return self.encode_headers() + self.encode_body()
	
	@classmethod
	def parse(cls, data, head=False):
		parser = HTTPParser(cls, head)
		parser.update(data)
		parser.eof()
		
		if parser.buffer:
			raise HTTPError("Got more data than expected")
		
		return parser.message


class HTTPRequest(HTTPMessage):
	def __init__(self):
		super().__init__()
		self.method = "GET"
		self.path = "/"
		
		self.params = None
		self.continue_threshold = 1024
	
	def check_path(self):
		if "?" in self.path:
			self.path, params = self.path.split("?", 1)
			self.params = formdecode(params)
		
	def encode_start_line(self):
		path = self.path
		if self.params is not None:
			path += "?" + formencode(self.params)
		return "%s %s %s" %(self.method, path, self.version)
	
	def parse_start_line(self, line):
		fields = line.split(maxsplit=2)
		if len(fields) != 3:
			raise HTTPError("Failed to parse HTTP request start line")
		
		self.method = fields[0]
		self.path = fields[1]
		self.version = fields[2]
		
		self.check_version()
		self.check_path()
		
	def encode_body(self):
		body = super().encode_body()
		if self.continue_threshold is not None:
			if len(body) > self.continue_threshold:
				self.headers["Expect"] = "100-continue"
		return body
	
	@classmethod
	def build(cls, method, path):
		params = None
		if "?" in path:
			path, data = path.split("?", 1)
			params = formdecode(data)
		
		request = cls()
		request.method = method
		request.path = path
		request.params = params
		return request
		
	@classmethod
	def get(cls, path):
		return cls.build("GET", path)
	@classmethod
	def post(cls, path):
		return cls.build("POST", path)
	@classmethod
	def put(cls, path):
		return cls.build("PUT", path)
	@classmethod
	def patch(cls, path):
		return cls.build("PATCH", path)
	@classmethod
	def delete(cls, path):
		return cls.build("DELETE", path)


class HTTPResponse(HTTPMessage):
	def __init__(self, status_code=500):
		super().__init__()
		self.status_code = status_code
		self.status_name = STATUS_NAMES.get(status_code, "Unknown")
		self.upgrade = None
	
	def success(self):
		return self.status_code // 100 == 2
		
	def error(self):
		return not self.success()
	
	def raise_if_error(self):
		if self.error():
			raise HTTPResponseError(self)
		
	def encode_start_line(self):
		return "%s %i %s" %(self.version, self.status_code, self.status_name)
		
	def parse_start_line(self, line):
		fields = line.split(maxsplit=2)
		if len(fields) not in [2, 3]:
			raise HTTPError("Failed to parse HTTP response start line")
		
		self.version = fields[0]
		self.check_version()
		
		if not fields[1].isdecimal():
			raise HTTPError("HTTP response has invalid status code")
		
		self.status_code = int(fields[1])
		
		self.status_name = ""
		if len(fields) == 3:
			self.status_name = fields[2]
		

class HTTPParser:
	def __init__(self, cls, head):
		self.message = cls()
		self.head = head
		
		self.buffer = b""
		self.state = self.state_header
	
	def complete(self): return self.state is None
	def header_complete(self): return self.state != self.state_header
	
	def update(self, data):
		self.buffer += data
		while not self.state():
			pass

	def eof(self):
		if self.state != self.state_body:
			raise HTTPError("Got unexpected EOF while parsing HTTP message")
		self.finish()
	
	def finish(self):
		self.message.parse_body()
		self.state = None
	
	def is_head(self):
		if self.head: return True
		if isinstance(self.message, HTTPResponse):	
			if self.message.status_code // 100 == 1: return True
			if self.message.status_code in [204, 304]: return True
		return False
	
	def state_header(self):
		if not b"\r\n\r\n" in self.buffer:
			return True
		
		header, self.buffer = self.buffer.split(b"\r\n\r\n", 1)
		
		try:
			lines = header.decode().split("\r\n")
		except UnicodeDecodeError:
			raise HTTPError("Failed to decode HTTP header")
			
		if len(lines) == 0:
			raise HTTPError("HTTP message must start with header line")
		
		self.message.parse_start_line(lines[0])
		
		for header in lines[1:]:
			if not ": " in header:
				raise HTTPError("Invalid line in HTTP headers")
			key, value = header.split(": ", 1)
			self.message.headers[key] = value
		
		if self.is_head():
			self.state = None
			return True
		
		if self.message.is_chunked():
			self.state = self.state_chunk_header
			return False
		elif "Content-Length" in self.message.headers:
			if not self.message.headers["Content-Length"].isdecimal():
				raise HTTPError("Invalid Content-Length header")
			self.state = self.state_fixed_body
			return False
		elif isinstance(self.message, HTTPResponse):
			self.state = self.state_body
			return False
		
		self.finish()
		return True
	
	def state_chunk_header(self):
		if not b"\r\n" in self.buffer:
			return True
			
		line, self.buffer = self.buffer.split(b"\r\n", 1)
		try:
			line = line.decode()
		except UnicodeDecodeError:
			raise HTTPError("Failed to decode chunk length")

		if not util.is_hexadecimal(line):
			raise HTTPError("Invalid HTTP chunk length")
		
		self.chunk_length = int(line, 16)
		
		self.state = self.state_chunk_body
		return False
		
	def state_chunk_body(self):
		if len(self.buffer) < self.chunk_length + 2:
			return True
			
		if self.buffer[self.chunk_length : self.chunk_length + 2] != b"\r\n":
			raise HTTPError("HTTP chunk should be terminated with \\r\\n")
		
		self.message.body += self.buffer[:self.chunk_length]
		
		self.buffer = self.buffer[self.chunk_length + 2:]
		
		if self.chunk_length == 0:
			self.finish()
			return True
			
		self.state = self.state_chunk_header
		return False
		
	def state_fixed_body(self):
		length = int(self.message.headers["Content-Length"])
		if len(self.buffer) < length:
			return True
			
		self.message.body = self.buffer[:length]
		self.buffer = self.buffer[length:]
		self.finish()
		return True
	
	def state_body(self):
		self.message.body += self.buffer
		self.buffer = b""
		return True


class HTTPClient:
	def __init__(self, sock):
		self.sock = sock
		self.buffer = b""
	
	async def send(self, data):
		await self.sock.send(data)
	
	async def recv(self):
		if self.buffer:
			buffer = self.buffer
			self.buffer = b""
			return buffer
		return await self.sock.recv()
	
	async def request(self, req, *, headerfunc=None, writefunc=None):
		logger.debug("Sending HTTP request headers")
		await self.send(req.encode_headers())
		
		if req.headers.get("Expect") == "100-continue":
			logger.debug("Waiting for 100-continue")
			response = await self.receive_response(True)
			if response.status_code != 100:
				raise HTTPError("Expected 100-continue response")
		
		body = req.encode_body()
		if body:
			logger.debug("Sending HTTP request body")
			await self.send(req.encode_body())
		
		response = await self.receive_response(req.method == "HEAD", headerfunc, writefunc)
		return response
			
	async def receive_response(self, head, headerfunc=None, writefunc=None):
		parser = HTTPParser(HTTPResponse, head)
		
		while not parser.header_complete():
			parser.update(await self.recv())
		
		if headerfunc:
			await headerfunc(parser.message)
		
		if writefunc:
			if parser.message.body:
				await writefunc(parser.message.body)
			offset = len(parser.message.body)
		
		while not parser.complete():
			try:
				parser.update(await self.recv())
			except util.StreamError:
				parser.eof()
			
			if writefunc and len(parser.message.body) > offset:
				await writefunc(parser.message.body[offset:])
				offset = len(parser.message.body)
		
		self.buffer += parser.buffer
		return parser.message
	
	def local_address(self):
		return self.sock.local_address()
	def remote_address(self):
		return self.sock.remote_address()
	def remote_certificate(self):
		return self.sock.remote_certificate()
		
		
class HTTPServerClient:
	def __init__(self, handler, client):
		self.handler = handler
		self.client = client
	
	async def process(self):
		try:
			parser = HTTPParser(HTTPRequest, False)
			while not parser.header_complete():
				data = await self.client.recv()
				parser.update(data)
			
			if parser.message.headers.get("Expect") == "100-continue":
				await self.client.send(HTTPResponse(100).encode())
			
			while not parser.complete():
				data = await self.client.recv()
				parser.update(data)
			
			response = await self.handle_request(parser.message)
			
			logger.info("Sending HTTP response (%i)", response.status_code)
			await self.client.send(response.encode())
			
			if response.upgrade:
				await response.upgrade()
		except Exception:
			logger.exception("Failed to process HTTP request")
	
	async def handle_request(self, request):
		logger.info("Received HTTP request: %s %s", request.method, request.path)
		
		try:
			response = await self.handler(self.client, request)
			if not isinstance(response, HTTPResponse):
				logger.error("HTTP handler must return HTTPResponse")
				response = HTTPResponse(500)
		except Exception:
			logger.exception("HTTP handler raised an exception")
			response = HTTPResponse(500)
		
		return response
		
		
class HTTPRoute:
	def __init__(self, router, path):
		self.router = router
		self.path = path
	
	def __enter__(self): return self
	def __exit__(self, typ, val, tb):
		self.router.remove(self.path)
		
		
class HTTPRouter:
	def __init__(self):
		self.routes = {}

	def route(self, path, handler):
		if path in self.routes:
			raise ValueError("Path is already routed")
		
		self.routes[path] = handler
		return HTTPRoute(self, path)
	
	def remove(self, path):
		del self.routes[path]
	
	async def handle(self, client, request):
		if request.path not in self.routes:
			logger.warning("HTTP router received unmapped request: %s" %request.path)
			return HTTPResponse(404)
		
		handler = self.routes[request.path]
		return await handler(client, request)
		

@contextlib.asynccontextmanager
async def connect(url, context=None):
	scheme, host, port, path = util.parse_url(url)
	
	if scheme == "http":
		context = None
	elif scheme == "https":
		if context is None:
			context = tls.TLSContext()
	elif scheme is not None:
		raise ValueError("Invalid HTTP url scheme: %s" %scheme)
	
	if path is not None:
		raise ValueError("URL must not contain a path")
	
	if port is None:
		port = 443 if context else 80
	
	logger.debug("Establishing HTTP connection with %s:%i", host, port)
	async with tls.connect(host, port, context) as client:
		yield HTTPClient(client)

async def request(url, req, context=None, **kwargs):
	logger.info("Performing HTTP request: %s %s", req.method, req.path)
	async with connect(url, context) as client:
		response = await client.request(req, **kwargs)
	logger.info("Received HTTP response: %i", response.status_code)
	return response

REQUEST_ARGS = [
	"headers", "body", "text", "files", "boundary", "form",
	"rawform", "json", "xml", "params", "continue_threshold"
]

async def call(url, method, **kwargs):
	scheme, host, port, path = util.parse_url(url)
	if path is None:
		path = "/"
	
	req = HTTPRequest.build(method, path)
	
	for var in REQUEST_ARGS:
		if var in kwargs:
			setattr(req, var, kwargs.pop(var))
	
	if "Host" not in req.headers:
		req.headers["Host"] = util.make_url(None, host, port, None)
	
	url = util.make_url(scheme, host, port, None)
	return await request(url, req, **kwargs)

async def get(url, **kwargs): return await call(url, "GET", **kwargs)
async def post(url, **kwargs): return await call(url, "POST", **kwargs)
async def put(url, **kwargs): return await call(url, "PUT", **kwargs)
async def patch(url, **kwargs): return await call(url, "PATCH", **kwargs)
async def delete(url, **kwargs): return await call(url, "DELETE", **kwargs)

@contextlib.asynccontextmanager
async def serve(handler, host="", port=0, context=None):
	async def handle(client):
		host, port = client.remote_address()
		logger.debug("New HTTP connection: %s:%i", host, port)
		
		client = HTTPServerClient(handler, client)
		await client.process()
	
	logger.info("Starting HTTP server at %s:%i", host, port)
	async with tls.serve(handle, host, port, context):
		yield
	logger.info("HTTP server is closed")

@contextlib.asynccontextmanager
async def serve_router(host="", port=0, context=None):
	router = HTTPRouter()
	async with serve(router.handle, host, port, context):
		yield router




from anynet import util
import contextlib
import itertools
import anyio
import time


class Scheduler:
	def __init__(self, group):
		self.group = group
		
		self.handle = itertools.count()
		self.event = anyio.Event()
		self.events = {}
		
	def start(self):
		self.group.start_soon(self.process)
	
	async def process(self):
		while True:
			timeout = self.process_timers()
			with anyio.move_on_after(timeout):
				await self.event.wait()
				self.event = anyio.Event()
	
	def process_timers(self):
		current = time.monotonic()
		items = self.events.copy().items()
		for handle, (deadline, repeat, function, args) in items:
			if deadline <= current:
				del self.events[handle]
				if repeat is not None:
					self.events[handle] = (deadline + repeat, repeat, function, args)
				self.group.start_soon(function, *args)
		
		timeouts = [event[0] - current for event in self.events.values()]
		return min(timeouts, default=None)
	
	def schedule(self, function, delay, *args):
		deadline = time.monotonic() + delay
		
		handle = next(self.handle)
		self.events[handle] = (deadline, None, function, args)
		self.event.set()
		return handle
	
	def repeat(self, function, delay, *args):
		deadline = time.monotonic() + delay
		
		handle = next(self.handle)
		self.events[handle] = (deadline, delay, function, args)
		self.event.set()
		return handle
		
	def remove(self, handle):
		if handle in self.events:
			del self.events[handle]
	
	def remove_all(self):
		self.events = {}


@contextlib.asynccontextmanager
async def create():
	async with util.create_task_group() as group:
		scheduler = Scheduler(group)
		scheduler.start()
		yield scheduler




import struct


class StreamOut:
	def __init__(self, endian):
		self.endian = endian
		self.data = bytearray()
		self.pos = 0
		self.stack = []
		
	def push(self): self.stack.append(self.pos)
	def pop(self): self.pos = self.stack.pop()
		
	def get(self): return bytes(self.data)
	def size(self): return len(self.data)
	def tell(self): return self.pos
	def seek(self, pos):
		if pos > len(self.data):
			self.data += bytes(pos - len(self.data))
		self.pos = pos
	def skip(self, num): self.seek(self.pos + num)
	def align(self, num): self.skip((num - self.pos % num) % num)
	def available(self): return len(self.data) - self.pos
	def eof(self): return self.pos >= len(self.data)
		
	def write(self, data):
		self.data[self.pos : self.pos + len(data)] = data
		self.pos += len(data)
		
	def pad(self, num, char=b"\0"):
		self.write(char * num)
		
	def ascii(self, data):
		self.write(data.encode("ascii"))
		
	def u8(self, value): self.write(bytes([value]))
	def u16(self, value): self.write(struct.pack(self.endian + "H", value))
	def u32(self, value): self.write(struct.pack(self.endian + "I", value))
	def u64(self, value): self.write(struct.pack(self.endian + "Q", value))
	
	def s8(self, value): self.write(struct.pack("b", value))
	def s16(self, value): self.write(struct.pack(self.endian + "h", value))
	def s32(self, value): self.write(struct.pack(self.endian + "i", value))
	def s64(self, value): self.write(struct.pack(self.endian + "q", value))
	
	def u24(self, value):
		if self.endian == ">":
			self.u16(value >> 8)
			self.u8(value & 0xFF)
		else:
			self.u8(value & 0xFF)
			self.u16(value >> 8)
			
	def float(self, value): self.write(struct.pack(self.endian + "f", value))
	def double(self, value): self.write(struct.pack(self.endian + "d", value))
	
	def bool(self, value): self.u8(1 if value else 0)
	def char(self, value): self.u8(ord(value))
	def wchar(self, value): self.u16(ord(value))
	
	def chars(self, data): self.repeat(data, self.char)
	def wchars(self, data): self.repeat(data, self.wchar)
	
	def repeat(self, list, func):
		for value in list:
			func(value)


class StreamIn:
	def __init__(self, data, endian):
		self.endian = endian
		self.data = data
		self.pos = 0
		self.stack = []
		
	def push(self): self.stack.append(self.pos)
	def pop(self): self.pos = self.stack.pop()
		
	def get(self): return self.data
	def size(self): return len(self.data)
	
	def tell(self): return self.pos
	def seek(self, pos):
		if pos > self.size():
			raise OverflowError("Buffer overflow")
		self.pos = pos
	
	def skip(self, num): self.seek(self.pos + num)
	def align(self, num): self.skip((num - self.pos % num) % num)
	def eof(self): return self.pos == len(self.data)
	def available(self): return len(self.data) - self.pos
	
	def peek(self, num):
		if self.available() < num:
			raise OverflowError("Buffer overflow")
		return self.data[self.pos : self.pos + num]
		
	def read(self, num):
		data = self.peek(num)
		self.skip(num)
		return data
		
	def readall(self):
		return self.read(self.available())
		
	def pad(self, num, char=b"\0"):
		if self.read(num) != char * num:
			raise ValueError("Incorrect padding")
			
	def ascii(self, num):
		return self.read(num).decode("ascii")
		
	def u8(self): return self.read(1)[0]
	def u16(self): return struct.unpack(self.endian + "H", self.read(2))[0]
	def u32(self): return struct.unpack(self.endian + "I", self.read(4))[0]
	def u64(self): return struct.unpack(self.endian + "Q", self.read(8))[0]
	
	def s8(self): return struct.unpack("b", self.read(1))[0]
	def s16(self): return struct.unpack(self.endian + "h", self.read(2))[0]
	def s32(self): return struct.unpack(self.endian + "i", self.read(4))[0]
	def s64(self): return struct.unpack(self.endian + "q", self.read(8))[0]
	
	def u24(self):
		if self.endian == ">":
			return (self.u16() << 8) | self.u8()
		return self.u8() | (self.u16() << 8)
	
	def float(self): return struct.unpack(self.endian + "f", self.read(4))[0]
	def double(self): return struct.unpack(self.endian + "d", self.read(8))[0]
	
	def bool(self): return bool(self.u8())
	def char(self): return chr(self.u8())
	def wchar(self): return chr(self.u16())
	
	def chars(self, num): return "".join(self.repeat(self.char, num))
	def wchars(self, num): return "".join(self.repeat(self.wchar, num))
	
	def repeat(self, func, count):
		return [func() for i in range(count)]
		
		
class BitStreamOut(StreamOut):
	def __init__(self, endian):
		super().__init__(endian)
		self.bitpos = 0
		
	def push(self): self.stack.append((self.pos, self.bitpos))
	def pop(self): self.pos, self.bitpos = self.stack.pop()
	
	def seek(self, pos):
		super().seek(pos)
		self.bitpos = 0
	
	def seekbits(self, pos):
		super().seek(pos // 8)
		self.bitpos = pos % 8
		if self.bitpos != 0 and self.eof():
			self.data += b"\0"
			
	def tellbits(self):
		return self.pos * 8 + self.bitpos
	
	def bytealign(self):
		if self.bitpos != 0:
			self.skip(1)
			self.bitpos = 0
			
	def align(self, num):
		self.bytealign()
		super().align(num)
		
	def bit(self, value):
		if self.pos == len(self.data):
			self.data += b"\0"

		byte = self.data[self.pos]
		mask = 1 << (7 - self.bitpos)
		if value:
			byte |= mask
		else:
			byte &= ~mask
		self.data[self.pos] = byte
		
		self.bitpos += 1
		if self.bitpos == 8:
			self.bitpos = 0
			self.pos += 1
			
	def bits(self, value, num):
		for i in range(num):
			self.bit((value >> (num - i - 1)) & 1)
			
	def write(self, data):
		if self.bitpos == 0: #Fast method
			super().write(data)
		else: #Slow method
			for value in data:
				self.bits(value, 8)
		
		
class BitStreamIn(StreamIn):
	def __init__(self, data, endian):
		super().__init__(data, endian)
		self.bitpos = 0
		
	def push(self): self.stack.append((self.pos, self.bitpos))
	def pop(self): self.pos, self.bitpos = self.stack.pop()
	
	def seek(self, pos):
		super().seek(pos)
		self.bitpos = 0
		
	def seekbits(self, pos):
		if pos > len(self.data) * 8:
			raise OverflowError("Buffer overflow")
		self.pos = pos // 8
		self.bitpos = pos % 8
		
	def tellbits(self):
		return self.pos * 8 + self.bitpos
		
	def bytealign(self):
		if self.bitpos != 0:
			self.skip(1)
			self.bitpos = 0
		
	def align(self, num):
		self.bytealign()
		super().align(num)
		
	def bit(self):
		if self.eof():
			raise OverflowError("Buffer overflow")
		
		byte = self.data[self.pos]
		value = (byte >> (7 - self.bitpos)) & 1
		
		self.bitpos += 1
		if self.bitpos == 8:
			self.bitpos = 0
			self.pos += 1
		
		return value
		
	def bits(self, num):
		value = 0
		for i in range(num):
			value = (value << 1) | self.bit()
		return value
		
	def read(self, num):
		if self.bitpos == 0: #Fast method
			return super().read(num)
		else: #Slow method
			data = []
			for i in range(num):
				data.append(self.bits(8))
			return bytes(data)




from anynet import tls, http, util, queue
import contextlib
import secrets
import hashlib
import base64
import struct
import anyio

import logging
logger = logging.getLogger(__name__)


OPCODE_CONTINUE = 0
OPCODE_TEXT = 1
OPCODE_BINARY = 2
OPCODE_DISCONNECT = 8
OPCODE_PING = 9
OPCODE_PONG = 10


def apply_mask(data, key):
	return bytes([data[i] ^ key[i % 4] for i in range(len(data))])

def calculate_key_hash(key):
	string = key.encode("ascii") + b"258EAFA5-E914-47DA-95CA-C5AB0DC85B11"
	hash = hashlib.sha1(string).digest()
	return base64.b64encode(hash).decode()

def check_handshake(request, protocol):
	if request.method != "GET": return 405
	
	if request.headers.get("Connection").lower() != "upgrade": return 400
	if request.headers.get("Upgrade") != "websocket": return 400
	
	if "Sec-WebSocket-Key" not in request.headers: return 400
	
	if protocol is not None:
		if "Sec-WebSocket-Protocol" not in request.headers:
			return 400
		
		protocols = request.headers["Sec-WebSocket-Protocol"].split(", ")
		if protocol not in protocols:
			return 400
	return 101

def accept_handshake(request, protocol):
	status = check_handshake(request, protocol)
	if status != 101:
		logger.info("WS handshake error: %i" %status)
		return http.HTTPResponse(status)
		
	accept = calculate_key_hash(request.headers["Sec-WebSocket-Key"])
	
	response = http.HTTPResponse(101)
	response.headers["Connection"] = "upgrade"
	response.headers["Upgrade"] = "WebSocket"
	response.headers["Sec-WebSocket-Accept"] = accept
	if protocol is not None:
		response.headers["Sec-WebSocket-Protocol"] = protocol
	
	logger.debug("WS handshake succeeded")
	return response


class WSError(Exception): pass


class WSPacket:
	def __init__(self, opcode, payload=b""):
		self.opcode = opcode
		self.payload = payload


class WSPacketClient:
	def __init__(self, client, group):
		self.client = client
		self.group = group
	
		self.packets = queue.create()
		
		self.buffer = b""
		self.fragments = None
		self.message_type = None # For continuation frames
		
		self.server_mode = False
	
	async def start_client(self, host, path, protocols):
		logger.debug("Performing WS handshake")
		
		self.server_mode = False
		
		key = secrets.token_urlsafe()
		
		request = http.HTTPRequest.get(path)
		request.headers["Host"] = host
		request.headers["Upgrade"] = "websocket"
		request.headers["Connection"] = "upgrade"
		request.headers["Sec-WebSocket-Key"] = key
		request.headers["Sec-WebSocket-Version"] = 13
		if protocols is not None:
			request.headers["Sec-WebSocket-Protocol"] = ", ".join(protocols)
		
		response = await self.client.request(request)
		if response.status_code != 101:
			raise WSError("WS server replied with status code %i" %response.status_code)
		
		if "Sec-WebSocket-Accept" not in response.headers:
			raise WSError("Sec-WebSocket-Accept header is missing")
		if response.headers["Sec-WebSocket-Accept"] != calculate_key_hash(key):
			raise WSError("Sec-WebSocket-Accept check failed")
		
		logger.debug("WS handshake succeeded")
		self.group.start_soon(self.process)
		
	async def start_server(self):
		self.server_mode = True
		
		self.group.start_soon(self.process)
	
	async def process(self):
		while True:
			await self.process_buffer()
			try:
				self.buffer += await self.client.recv()
			except util.StreamError:
				logger.debug("WS: underlying connection was closed")
				await self.packets.close()
				return
			
	async def process_buffer(self):
		while self.buffer:
			if len(self.buffer) < 2: return
			
			fin = self.buffer[0] >> 7
			opcode = self.buffer[0] & 0xF
			mask = self.buffer[1] >> 7
			size = self.buffer[1] & 0x7F
			
			offset = 2
			if size == 126:
				if len(self.buffer) < offset + 2: return
				size = struct.unpack_from(">H", self.buffer, offset)[0]
				offset += 2
			elif size == 127:
				if len(self.buffer) < offset + 8: return
				size = struct.unpack_from(">Q", self.buffer, 2)[0]
				offset += 8

			mask_key = b"\0\0\0\0"
			if mask:
				if len(self.buffer) < offset + 4: return
				mask_key = self.buffer[offset : offset + 4]
				offset += 4
				
			if len(self.buffer) < offset + size: return
			payload = apply_mask(self.buffer[offset : offset + size], mask_key)
			
			self.buffer = self.buffer[offset + size:]
			
			await self.process_packet(opcode, payload, fin)
			
	async def process_packet(self, opcode, payload, fin):
		if opcode in [OPCODE_TEXT, OPCODE_BINARY, OPCODE_CONTINUE]:
			if opcode in [OPCODE_TEXT, OPCODE_BINARY]:
				if self.message_type is not None:
					raise WSError("Expected continuation frame")
				self.message_type = opcode
				self.fragments = payload
			else:
				if self.message_type is None:
					raise WSError("Received unexpected continuation frame")
				self.fragments += payload
			
			if fin:
				packet = WSPacket(self.message_type, self.fragments)
				self.message_type = None
				self.fragments = None
				await self.packets.put(packet)
		else:
			if not fin:
				raise WSError("Control frame must have FIN set")
			packet = WSPacket(opcode, payload)
			await self.packets.put(packet)
			
	async def send(self, opcode, payload=b""):
		data = bytes([0x80 | opcode])
		
		mask = 0x80 if not self.server_mode else 0
		
		length = len(payload)
		if length < 126:
			data += bytes([mask | length])
		elif length <= 0xFFFF:
			data += struct.pack(">BH", mask | 0x7E, length)
		else:
			data += struct.pack(">BQ", mask | 0x7F, length)
			
		if not self.server_mode:
			mask = secrets.token_bytes(4)
			payload = mask + apply_mask(payload, mask)
		data += payload

		await self.client.send(data)
		
	async def recv(self):
		return await self.packets.get()
	
	def local_address(self): return self.client.local_address()
	def remote_address(self): return self.client.remote_address()
	def remote_certificate(self): return self.client.remote_certificate()


class WebSocketClient:
	def __init__(self, client, group):
		self.client = WSPacketClient(client, group)
		self.group = group
		
		self.binary_packets = queue.create()
		self.text_packets = queue.create()
	
	async def __aenter__(self): return self
	async def __aexit__(self, typ, exc, tb):
		await self.stop()
		
	async def start_client(self, host, path, protocols):
		await self.client.start_client(host, path, protocols)
		self.group.start_soon(self.process)
		
	async def start_server(self):
		await self.client.start_server()
		self.group.start_soon(self.process)
		
	async def process(self):
		while True:
			try:
				packet = await self.client.recv()
			except util.StreamError:
				await self.stop()
				return
			await self.process_packet(packet.opcode, packet.payload)
	
	async def process_packet(self, opcode, payload):
		if opcode == OPCODE_BINARY:
			await self.binary_packets.put(payload)
		elif opcode == OPCODE_TEXT:
			await self.text_packets.put(payload.decode())
		elif opcode == OPCODE_PING:
			await self.client.send(OPCODE_PONG, payload)
		elif opcode == OPCODE_DISCONNECT:
			logger.debug("Received disconnect packet")
			await self.client.send(OPCODE_DISCONNECT)
			await self.stop()
		else:
			raise ValueError("WS packet has unknown opcode: %i" %opcode)
	
	async def stop(self):
		await self.binary_packets.close()
		await self.text_packets.close()
		
	async def close(self):
		logger.debug("Closing WS connection")
		await self.stop()
		with util.catch(util.StreamError, silent=True):
			await self.client.send(OPCODE_DISCONNECT)
		logger.debug("WS connection is closed")
			
	async def send(self, data):
		await self.client.send(OPCODE_BINARY, data)
	async def send_text(self, text):
		await self.client.send(OPCODE_TEXT, text.encode())

	async def recv(self):
		return await self.binary_packets.get()
	async def recv_text(self):
		return await self.text_packets.get()
		
	def local_address(self): return self.client.local_address()
	def remote_address(self): return self.client.remote_address()
	def remote_certificate(self): return self.client.remote_certificate()


@contextlib.asynccontextmanager
async def connect(url, context=None, *, protocols=None):
	logger.debug("Connecting WS client to %s", url)
	
	scheme, host, port, path = util.parse_url(url)
		
	if scheme == "ws":
		context = None
	elif scheme == "wss":
		if context is None:
			context = tls.TLSContext()
	elif scheme is not None:
		raise ValueError("Invalid WS url scheme: %s" %scheme)
	
	if path is None:
		path = "/"
	
	server = util.make_url(None, host, port, None)
	async with http.connect(server, context) as client:
		async with util.create_task_group() as group:
			async with WebSocketClient(client, group) as client:
				await client.start_client(host, path, protocols)
				yield client
				await client.close()
	
	logger.debug("WS client is closed")

@contextlib.asynccontextmanager
async def route(handler, router, path, *, protocol=None):
	async def process(client):
		async with util.create_task_group() as group:
			async with WebSocketClient(client, group) as client:
				await client.start_server()
				await handler(client)
				await client.close()
	
	async def handle(client, request):
		host, port = client.remote_address()
		logger.debug("New WS connection: %s:%i", host, port)
		
		response = accept_handshake(request, protocol)
		if response.status_code == 101:
			response.upgrade = lambda: process(client)
		return response
	
	logger.info("Routing WS server at %s" %path)
	with router.route(path, handle):
		yield
	logger.info("WS server is closed")

@contextlib.asynccontextmanager
async def serve(handler, host="", port=0, context=None, *, path="/", protocol=None):
	async with http.serve_router(host, port, context) as router:
		async with route(handler, router, path, protocol=protocol):
			yield




from anynet import util
import contextlib
import socket
import anyio

import logging
logger = logging.getLogger(__name__)


class UDPSocket:
	def __init__(self, sock):
		self.sock = sock
		self.lock = anyio.Lock()
	
	async def send(self, data, addr):
		async with self.lock:
			await self.sock.sendto(data, addr[0], addr[1])
	async def recv(self):
		return await self.sock.receive()
	
	async def broadcast(self, data, port):
		host = util.broadcast_address()
		await self.send(data, (host, port))
	
	def local_address(self):
		return self.sock.extra(anyio.abc.SocketAttribute.local_address)


class UDPClient:
	def __init__(self, sock):
		self.sock = sock
		self.lock = anyio.Lock()
	
	async def send(self, data):
		async with self.lock:
			await self.sock.send(data)
	async def recv(self):
		return await self.sock.receive()
	
	def local_address(self):
		return self.sock.extra(anyio.abc.SocketAttribute.local_address)
	def remote_address(self):
		return self.sock.extra(anyio.abc.SocketAttribute.remote_address)


@contextlib.asynccontextmanager
async def bind(host="", port=0):
	if not host:
		host = util.local_address()
	
	logger.debug("Creating UDP socket at %s:%i", host, port)
	
	sock = await anyio.create_udp_socket(local_host=host, local_port=port)
	async with sock:
		rawsock = sock.extra(anyio.abc.SocketAttribute.raw_socket)
		rawsock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, True)
		yield UDPSocket(sock)

@contextlib.asynccontextmanager
async def connect(host, port):
	logger.debug("Connecting UDP client to %s:%i", host, port)
	
	sock = await anyio.create_connected_udp_socket(host, port)
	async with sock:
		yield UDPClient(sock)




import setuptools

long_description = \
	"This is a simple asynchronous networking library that " \
	"I'm using for hobby projects. It's based on anyio, so " \
	"it can be used together with either asyncio or trio."

setuptools.setup(
	name = "anynet",
	version = "0.0.13",
	description = "Networking library based on anyio",
	long_description = long_description,
	author = "Yannik Marchand",
	author_email = "ymarchand@me.com",
	url = "https://github.com/kinnay/anynet",
	packages = ["anynet"],
	license = "MIT",
	
	install_requires=[
		"anyio >= 3.1.0",
		"cryptography",
		"pyopenssl",
		"netifaces",
		"pycryptodome",
		"multidict"
	]
)




import contextlib
import netifaces
import struct
import socket
import string
import anyio
import math

import logging
logger = logging.getLogger(__name__)


StreamError = (
	anyio.EndOfStream, anyio.ClosedResourceError,
	anyio.BrokenResourceError, BrokenPipeError
)


def is_decimal(s):
	return s.isdecimal()

def is_hexadecimal(s):
	return s and all(c in string.hexdigits for c in s)

def ip_to_hex(ip):
	try:
		data = socket.inet_aton(ip)
	except OSError:
		raise ValueError("IP address is invalid")
	return struct.unpack(">I", data)[0]

def ip_from_hex(value):
	return socket.inet_ntoa(struct.pack(">I", value))
	
def local_address():
	interface = netifaces.gateways()["default"][netifaces.AF_INET][1]
	addresses = netifaces.ifaddresses(interface)[netifaces.AF_INET][0]
	return addresses["addr"]
	
def broadcast_address():
	interface = netifaces.gateways()["default"][netifaces.AF_INET][1]
	addresses = netifaces.ifaddresses(interface)[netifaces.AF_INET][0]
	return addresses["broadcast"]

def parse_url(url):
	scheme = None
	if "://" in url:
		scheme, url = url.split("://", 1)
	
	path = None
	if "/" in url:
		url, path = url.split("/", 1)
		path = "/" + path
	
	host = url
	port = None
	if ":" in url:
		host, port = url.split(":", 1)
		port = int(port)
	
	return scheme, host, port, path

def make_url(scheme, host, port, path):
	url = ""
	if scheme is not None:
		url += scheme + "://"
	url += host
	if port is not None:
		url += ":%i" %port
	if path is not None:
		if not path.startswith("/"):
			raise ValueError("Path must start with '/'")
		url += path
	return url

@contextlib.asynccontextmanager
async def create_task_group():
	async with anyio.create_task_group() as group:
		yield group
		group.cancel_scope.cancel()

@contextlib.contextmanager
def catch(cls, *, silent=False):
	try:
		yield
	except anyio.ExceptionGroup as e:
		filtered = []
		for exc in e.exceptions:
			if isinstance(exc, cls):
				if not silent:
					logger.exception("An exception occurred")
			else:
				filtered.append(exc)
		e.exceptions = filtered
		if filtered:
			raise
	except cls as e:
		if not silent:
			logger.exception("An exception occurred")
